package tk.eclipse.plugin.jseditor.editors;

import java.io.FileInputStream;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.jface.text.IRegion;
import org.eclipse.jface.text.ITextViewer;
import org.eclipse.jface.text.contentassist.CompletionProposal;
import org.eclipse.jface.text.contentassist.ContextInformation;
import org.eclipse.jface.text.contentassist.ContextInformationValidator;
import org.eclipse.jface.text.contentassist.ICompletionProposal;
import org.eclipse.jface.text.contentassist.IContextInformation;
import org.eclipse.jface.text.contentassist.IContextInformationValidator;
import org.eclipse.jface.text.templates.TemplateContextType;
import org.eclipse.swt.graphics.Image;

import tk.eclipse.plugin.htmleditor.HTMLPlugin;
import tk.eclipse.plugin.htmleditor.HTMLProjectParams;
import tk.eclipse.plugin.htmleditor.HTMLUtil;
import tk.eclipse.plugin.htmleditor.IOUtil;
import tk.eclipse.plugin.htmleditor.template.HTMLTemplateAssistProcessor;
import tk.eclipse.plugin.htmleditor.template.HTMLTemplateManager;
import tk.eclipse.plugin.htmleditor.template.JavaScriptContextType;
import tk.eclipse.plugin.jseditor.editors.additional.AdditionalJavaScriptCompleterManager;
import tk.eclipse.plugin.jseditor.editors.additional.IAdditionalJavaScriptCompleter;
import tk.eclipse.plugin.jseditor.launch.JavaScriptLibraryTable;

/**
 * IContentAssistProcessor implementation for JavaScriptEditor.
 * 
 * @author Naoki Takezoe
 */
public class JavaScriptAssistProcessor extends HTMLTemplateAssistProcessor { /* implements IContentAssistProcessor {*/
	
	private static List<AssistInfo> staticAssistInfo = new ArrayList<AssistInfo>();
	
	// assist proposal types
	public static final int VARIABLE =  0;
	public static final int FUNCTION =  1;
	public static final int KEYWORD  =  2;
	public static final int OBJECT   =  3;
	public static final int STATIC_VARIABLE = 4;
	public static final int STATIC_FUNCTION = 5;
	public static final int UNDEF    = 99;
	
	private static final String[] GLOBAL_PROPERTIES = {
			"Infinity", "NaN", "undefined"
	};
	
	private static final String[] GLOBAL_FUNCTIONS = {
			"decodeURI()", "decodeURIComponent()",
			"encodeURI()", "encodeURIComponent()",
			"escape()", "unescape()", "eval()", "isNaN()",
			"parseFloat()", "parseInt()", "taint()", "untaint()",
			"alert()"
	};
	
	private static final String[] CLASSES = {
			"Arguments", "Array", "Boolean", "Date", "Error",
			"Function", "Math", "NativeError", "Number", "Object", "RegExp", "String"
	};
	
	private static Map<String, AssistInfo[]> STATIC_MEMBERS = new HashMap<String, AssistInfo[]>();
	
	static {
//		// add keyword to static assist informations
//		for(int i=0;i<JavaScriptScanner.KEYWORDS.length;i++){
//			staticAssistInfo.add(new AssistInfo(KEYWORD, JavaScriptScanner.KEYWORDS[i]));
//		}
		staticAssistInfo.add(new AssistInfo(VARIABLE, "window"));
		staticAssistInfo.add(new AssistInfo(VARIABLE, "document"));
		staticAssistInfo.add(new AssistInfo(VARIABLE, "navigator"));
		staticAssistInfo.add(new AssistInfo(VARIABLE, "history"));
		staticAssistInfo.add(new AssistInfo(VARIABLE, "screen"));
		staticAssistInfo.add(new AssistInfo(VARIABLE, "location"));
		
		for(int i=0;i<GLOBAL_PROPERTIES.length;i++){
			staticAssistInfo.add(new AssistInfo(VARIABLE, GLOBAL_PROPERTIES[i], GLOBAL_PROPERTIES[i].length()));
		}
		for(int i=0;i<GLOBAL_FUNCTIONS.length;i++){
			staticAssistInfo.add(new AssistInfo(FUNCTION, GLOBAL_FUNCTIONS[i], GLOBAL_FUNCTIONS[i].length()-1));
		}
		for(int i=0;i<CLASSES.length;i++){
			staticAssistInfo.add(new AssistInfo(OBJECT, CLASSES[i], CLASSES[i], CLASSES[i].length()));
		}
		
		AssistInfo[] math = {
				new AssistInfo(VARIABLE, "E - Math", "E"),
				new AssistInfo(VARIABLE, "LN10 - Math", "LN10"),
				new AssistInfo(VARIABLE, "LN2 - Math", "LN2"),
				new AssistInfo(VARIABLE, "LOG10E - Math", "LOG10E"),
				new AssistInfo(VARIABLE, "LOG2E - Math", "LOG2E"),
				new AssistInfo(VARIABLE, "PI - Math", "PI"),
				new AssistInfo(VARIABLE, "SQRT1_2 - Math", "SQRT1_2"),
				new AssistInfo(VARIABLE, "SQRT2 - Math", "SQRT2"),
				new AssistInfo(FUNCTION, "abs() - Math", "abs()"),
				new AssistInfo(FUNCTION, "acos() - Math", "acos()"),
				new AssistInfo(FUNCTION, "asin() - Math", "asin()"),
				new AssistInfo(FUNCTION, "atan() - Math", "atan()"),
				new AssistInfo(FUNCTION, "atan2() - Math", "atan2()"),
				new AssistInfo(FUNCTION, "ceil() - Math", "ceil()"),
				new AssistInfo(FUNCTION, "cos() - Math", "cos()"),
				new AssistInfo(FUNCTION, "exp() - Math", "exp()"),
				new AssistInfo(FUNCTION, "floor() - Math", "floor()"),
				new AssistInfo(FUNCTION, "log() - Math", "log()"),
				new AssistInfo(FUNCTION, "max() - Math", "max()"),
				new AssistInfo(FUNCTION, "min() - Math", "min()"),
				new AssistInfo(FUNCTION, "pow() - Math", "pow()"),
				new AssistInfo(FUNCTION, "random() - Math", "random()"),
				new AssistInfo(FUNCTION, "round() - Math", "round()"),
				new AssistInfo(FUNCTION, "sin() - Math", "sin()"),
				new AssistInfo(FUNCTION, "sqrt() - Math", "sqrt()"),
				new AssistInfo(FUNCTION, "tan() - Math", "tan()"),
		};
		STATIC_MEMBERS.put("Math", math);
		
		AssistInfo[] object = {
				new AssistInfo(VARIABLE, "constructor - Object", "constructor"),
				new AssistInfo(VARIABLE, "prototype - Object", "prototype"),
				new AssistInfo(FUNCTION, "hasOwnProperty() - Object", "hasOwnProperty()"),
//				new AssistInfo(FUNCTION, "eval() - Object", "eval()"),
				new AssistInfo(FUNCTION, "hasOwnProperty() - Object", "hasOwnProperty()"),
				new AssistInfo(FUNCTION, "isPrototyprOf() - Object", "isPrototyprOf()"),
				new AssistInfo(FUNCTION, "propertyIsEnumerable() - Object", "propertyIsEnumerable()"),
				new AssistInfo(FUNCTION, "toLocaleString() - Object", "toLocaleString()"),
				new AssistInfo(FUNCTION, "toSource() - Object", "toSource()"),
				new AssistInfo(FUNCTION, "toString() - Object", "toString()"),
				new AssistInfo(FUNCTION, "unwatch() - Object", "unwatch()"),
				new AssistInfo(FUNCTION, "valueOf() - Object", "valueOf()"),
				new AssistInfo(FUNCTION, "watch() - Object","watch()"),
		};
		STATIC_MEMBERS.put("Object", object);
		
		AssistInfo[] element = {
				new AssistInfo(VARIABLE, "style - Element", "style"),
				new AssistInfo(VARIABLE, "firstChild - Element", "firstChild"),
				new AssistInfo(VARIABLE, "lastChild - Element", "lastChild"),
				new AssistInfo(VARIABLE, "id - Element", "id"),
				new AssistInfo(VARIABLE, "className - Element", "className"),
				new AssistInfo(VARIABLE, "nextSibling - Element", "nextSibling"),
				new AssistInfo(VARIABLE, "offsetHeight - Element", "offsetHeight"),
				new AssistInfo(VARIABLE, "offsetLeft - Element", "offsetLeft"),
				new AssistInfo(VARIABLE, "offsetTop - Element", "offsetTop"),
				new AssistInfo(VARIABLE, "offsetWidth - Element", "offsetWidth"),
				new AssistInfo(VARIABLE, "parentNode - Element", "parentNode"),
				new AssistInfo(VARIABLE, "previousSibling - Element", "previousSibling"),
				new AssistInfo(FUNCTION, "appendChild() - Element", "appendChild()"),
				new AssistInfo(FUNCTION, "cloneNode() - Element", "cloneNode()"),
				new AssistInfo(FUNCTION, "getAttribute() - Element", "getAttribute()"),
				new AssistInfo(FUNCTION, "hasChildNodes() - Element", "hasChildNodes()"),
				new AssistInfo(FUNCTION, "insertBefore() - Element", "insertBefore()"),
				new AssistInfo(FUNCTION, "removeAttribute() - Element", "removeAttribute()"),
				new AssistInfo(FUNCTION, "removeChild() - Element", "removeChild()"),
				new AssistInfo(FUNCTION, "setAttribute() - Element", "setAttribute()"),
		};
		STATIC_MEMBERS.put("Element", element);
		
		AssistInfo[] document = {
				new AssistInfo(FUNCTION, "createElement() - document", "createElement()"),
				new AssistInfo(FUNCTION, "getElementById() - document", "getElementById()"),
				new AssistInfo(FUNCTION, "getElementsByName() - document", "getElementsByName()"),
				new AssistInfo(FUNCTION, "getElementsByTagName() - document", "getElementsByTagName()"),
				new AssistInfo(VARIABLE, "alinkColor - document", "alinkColor"),
				new AssistInfo(VARIABLE, "bgColor - document", "bgColor"),
				new AssistInfo(VARIABLE, "cookie - document", "cookie"),
				new AssistInfo(VARIABLE, "defaultCharset - document", "defaultCharset"),
				new AssistInfo(VARIABLE, "domain - document", "domain"),
				new AssistInfo(VARIABLE, "fgColor - document", "fgColor"),
				new AssistInfo(VARIABLE, "fileCreatedDate - document", "fileCreatedDate"),
				new AssistInfo(VARIABLE, "fileModifiedDate - document", "fileModifiedDate"),
				new AssistInfo(VARIABLE, "fileSize - document", "fileSize"),
				new AssistInfo(VARIABLE, "lastMofified - document", "lastMofified"),
				new AssistInfo(VARIABLE, "linkColor - document", "linkColor"),
				new AssistInfo(VARIABLE, "readyState - document", "readyState"),
				new AssistInfo(VARIABLE, "referrer - document", "referrer"),
				new AssistInfo(VARIABLE, "title - document", "title"),
				new AssistInfo(VARIABLE, "URL - document", "URL"),
				new AssistInfo(VARIABLE, "vlinkColor - document", "vlinkColor"),
				new AssistInfo(FUNCTION, "clear() - document", "clear()"),
				new AssistInfo(FUNCTION, "close() - document", "close()"),
				new AssistInfo(FUNCTION, "getSelection() - document", "getSelection()"),
				new AssistInfo(FUNCTION, "open() - document", "open()"),
				new AssistInfo(FUNCTION, "routeEvent() - document", "routeEvent()"),
				new AssistInfo(FUNCTION, "write() - document", "write()"),
				new AssistInfo(FUNCTION, "writeln() - document", "writeln()"),
		};
		STATIC_MEMBERS.put("document", document);
		
		AssistInfo[] style = {
				new AssistInfo(VARIABLE, "background"),
				new AssistInfo(VARIABLE, "backgroundAttachment"),
				new AssistInfo(VARIABLE, "backgroundColor"),
				new AssistInfo(VARIABLE, "backgroundImage"),
				new AssistInfo(VARIABLE, "backgroundPosition"),
				new AssistInfo(VARIABLE, "backgroundRepeat"),
				new AssistInfo(VARIABLE, "border"),
				new AssistInfo(VARIABLE, "borderBottom"),
				new AssistInfo(VARIABLE, "borderBottomColor"),
				new AssistInfo(VARIABLE, "borderBottomStyle"),
				new AssistInfo(VARIABLE, "borderBottomWidth"),
				new AssistInfo(VARIABLE, "borderColor"),
				new AssistInfo(VARIABLE, "borderLeft"),
				new AssistInfo(VARIABLE, "borderLeftColor"),
				new AssistInfo(VARIABLE, "borderLeftStyle"),
				new AssistInfo(VARIABLE, "borderLeftWidth"),
				new AssistInfo(VARIABLE, "borderRight"),
				new AssistInfo(VARIABLE, "borderRightColor"),
				new AssistInfo(VARIABLE, "borderRightStyle"),
				new AssistInfo(VARIABLE, "borderRightWidth"),
				new AssistInfo(VARIABLE, "borderStyle"),
				new AssistInfo(VARIABLE, "borderTop"),
				new AssistInfo(VARIABLE, "borderTopColor"),
				new AssistInfo(VARIABLE, "borderTopStyle"),
				new AssistInfo(VARIABLE, "borderTopWidth"),
				new AssistInfo(VARIABLE, "clear"),
				new AssistInfo(VARIABLE, "clip"),
				new AssistInfo(VARIABLE, "color"),
				new AssistInfo(VARIABLE, "cursor"),
				new AssistInfo(VARIABLE, "display"),
				new AssistInfo(VARIABLE, "filter"),
				new AssistInfo(VARIABLE, "font"),
				new AssistInfo(VARIABLE, "fontFamily"),
				new AssistInfo(VARIABLE, "fontSize"),
				new AssistInfo(VARIABLE, "fontVariant"),
				new AssistInfo(VARIABLE, "fontWeight"),
				new AssistInfo(VARIABLE, "height"),
				new AssistInfo(VARIABLE, "left"),
				new AssistInfo(VARIABLE, "letterSpacing"),
				new AssistInfo(VARIABLE, "lineHeight"),
				new AssistInfo(VARIABLE, "lineStyle"),
				new AssistInfo(VARIABLE, "lineStyleImage"),
				new AssistInfo(VARIABLE, "lineStylePosition"),
				new AssistInfo(VARIABLE, "lineStyleType"),
				new AssistInfo(VARIABLE, "margin"),
				new AssistInfo(VARIABLE, "marginBottom"),
				new AssistInfo(VARIABLE, "marginLeft"),
				new AssistInfo(VARIABLE, "marginRight"),
				new AssistInfo(VARIABLE, "marginTop"),
				new AssistInfo(VARIABLE, "overflow"),
				new AssistInfo(VARIABLE, "padding"),
				new AssistInfo(VARIABLE, "paddingBottom"),
				new AssistInfo(VARIABLE, "paddingLeft"),
				new AssistInfo(VARIABLE, "paddingRight"),
				new AssistInfo(VARIABLE, "paddingTop"),
				new AssistInfo(VARIABLE, "pageBreakAfter"),
				new AssistInfo(VARIABLE, "pageBreakBefore"),
				new AssistInfo(VARIABLE, "position"),
				new AssistInfo(VARIABLE, "styleFloat"),
				new AssistInfo(VARIABLE, "textAlign"),
				new AssistInfo(VARIABLE, "textDecoration"),
				new AssistInfo(VARIABLE, "textDecorationBlink"),
				new AssistInfo(VARIABLE, "textDecorationLineThrough"),
				new AssistInfo(VARIABLE, "textDecorationNone"),
				new AssistInfo(VARIABLE, "textDecorationOverline"),
				new AssistInfo(VARIABLE, "textDecorationUnderline"),
				new AssistInfo(VARIABLE, "textIndent"),
				new AssistInfo(VARIABLE, "textTransform"),
				new AssistInfo(VARIABLE, "top"),
				new AssistInfo(VARIABLE, "verticalAlign"),
				new AssistInfo(VARIABLE, "visibility"),
				new AssistInfo(VARIABLE, "whiteSpace"),
				new AssistInfo(VARIABLE, "width"),
				new AssistInfo(VARIABLE, "wordBreak"),
				new AssistInfo(VARIABLE, "wordWrap"),
				new AssistInfo(VARIABLE, "zIndex"),
		};
		STATIC_MEMBERS.put("style", style);
		
		AssistInfo[] window = {
				new AssistInfo(VARIABLE, "location - window", "location"),
				new AssistInfo(VARIABLE, "history - window", "history"),
				new AssistInfo(VARIABLE, "document - window", "document"),
		};
		STATIC_MEMBERS.put("window", window);
		
		AssistInfo[] history = {
				new AssistInfo(VARIABLE, "length - history", "length"),
				new AssistInfo(FUNCTION, "back() - history", "back()"),
				new AssistInfo(FUNCTION, "forward() - history", "forward()"),
				new AssistInfo(FUNCTION, "go() - history", "go()"),
		};
		STATIC_MEMBERS.put("history", history);
		
		AssistInfo[] location = {
				new AssistInfo(VARIABLE, "hash - location", "hash"),
				new AssistInfo(VARIABLE, "host - location", "host"),
				new AssistInfo(VARIABLE, "hostname - location", "hostname"),
				new AssistInfo(VARIABLE, "href - location", "href"),
				new AssistInfo(VARIABLE, "port - location", "port"),
				new AssistInfo(VARIABLE, "protocol - location", "protocol"),
				new AssistInfo(VARIABLE, "search - location", "search"),
				new AssistInfo(FUNCTION, "reload() - location", "reload()"),
				new AssistInfo(FUNCTION, "replace() - location", "replace()")
		};
		STATIC_MEMBERS.put("location", location);
		
		AssistInfo[] screen = {
				new AssistInfo(VARIABLE, "colorDepth - screen", "colorDepth"),
				new AssistInfo(VARIABLE, "fontSmoothingEnabled - screen", "fontSmoothingEnabled"),
				new AssistInfo(VARIABLE, "height - screen", "height"),
				new AssistInfo(VARIABLE, "updateInterval - screen", "updateInterval"),
				new AssistInfo(VARIABLE, "width - screen", "width"),
		};
		STATIC_MEMBERS.put("screen", screen);
		
		AssistInfo[] date = {
				new AssistInfo(FUNCTION, "getFullYear() - Date", "getFullYear()"),
				new AssistInfo(FUNCTION, "setFullYear(year) - Date", "setFullYear()"),
				new AssistInfo(FUNCTION, "getYear() - Date", "getYear()"),
				new AssistInfo(FUNCTION, "setYear(year) - Date", "setYear()"),
				new AssistInfo(FUNCTION, "getMonth() - Date", "getMonth()"),
				new AssistInfo(FUNCTION, "setMonth(mon) - Date", "setMonth()"),
				new AssistInfo(FUNCTION, "getDate() - Date", "getDate()"),
				new AssistInfo(FUNCTION, "setDate(date) - Date", "setDate()"),
				new AssistInfo(FUNCTION, "getHours() - Date", "getHours()"),
				new AssistInfo(FUNCTION, "setHours(hour) - Date", "setHours()"),
				new AssistInfo(FUNCTION, "getMinutes() - Date", "getMinutes()"),
				new AssistInfo(FUNCTION, "setMinutes(min) - Date", "setMinutes()"),
				new AssistInfo(FUNCTION, "getSeconds() - Date", "getSeconds()"),
				new AssistInfo(FUNCTION, "setSeconds(sec) - Date", "setSeconds()"),
				new AssistInfo(FUNCTION, "getMilliseconds() - Date", "getMilliseconds()"),
				new AssistInfo(FUNCTION, "setMilliseconds(msec) - Date", "setMilliseconds()"),
				new AssistInfo(FUNCTION, "getTimezoneOffset() - Date", "getTimezoneOffset()"),
				new AssistInfo(FUNCTION, "toString() - Date", "toString()"),
				new AssistInfo(FUNCTION, "toLocaleString() - Date", "toLocaleString()"),
				new AssistInfo(FUNCTION, "toGMTString() - Date", "toGMTString()"),
				new AssistInfo(FUNCTION, "toUTCString() - Date", "toUTCString()"),
				new AssistInfo(FUNCTION, "toVarDate() - Date", "toVarDate()"),
				new AssistInfo(FUNCTION, "getTime() - Date", "getTime()"),
				new AssistInfo(FUNCTION, "setTime(time) - Date", "setTime()"),
				new AssistInfo(FUNCTION, "UTC(year, mon, day, hour, min, sec) - Date", "UTC()"),
				new AssistInfo(FUNCTION, "parse(str) - Date", "parse()"),
		};
		STATIC_MEMBERS.put("Date", date);
		
		AssistInfo[] string = {
				new AssistInfo(VARIABLE, "length - String", "length"),
				new AssistInfo(FUNCTION, "charAt(n) - String", "charAt()"),
				new AssistInfo(FUNCTION, "substring(from[, to]) - String", "substring()"),
				new AssistInfo(FUNCTION, "slice(from[, to]) - String", "slise()"),
				new AssistInfo(FUNCTION, "substr(from[, to]) - String", "substr()"),
				new AssistInfo(FUNCTION, "split([separator[, limit]]) - String", "split()"),
				new AssistInfo(FUNCTION, "concat(str) - String", "concat()"),
				new AssistInfo(FUNCTION, "replace(regexp, newstr) - String", "replace()"),
				new AssistInfo(FUNCTION, "toUpperCase() - String", "toUpperCase()"),
				new AssistInfo(FUNCTION, "toLowerCase() - String", "toLowerCase()"),
				new AssistInfo(FUNCTION, "indexOf(keyword[, from]) - String", "indexOf()"),
				new AssistInfo(FUNCTION, "lastIndexOf(keyword[, from]) - String", "lastIndexOf()"),
				new AssistInfo(FUNCTION, "match(regexp) - String", "match()"),
				new AssistInfo(FUNCTION, "search(regexp) - String", "search()"),
				new AssistInfo(FUNCTION, "charCodeAt(n) - String", "charCodeAt()"),
				new AssistInfo(FUNCTION, "bold() - String", "bold()"),
				new AssistInfo(FUNCTION, "italics() - String", "italics()"),
				new AssistInfo(FUNCTION, "fixed() - String", "fixed()"),
				new AssistInfo(FUNCTION, "big() - String", "big()"),
				new AssistInfo(FUNCTION, "small() - String", "small()"),
				new AssistInfo(FUNCTION, "blink() - String", "blink()"),
				new AssistInfo(FUNCTION, "strike() - String", "strike()"),
				new AssistInfo(FUNCTION, "sup() - String", "sup()"),
				new AssistInfo(FUNCTION, "sub() - String", "sub()"),
				new AssistInfo(FUNCTION, "fontcolor(color) - String", "fontcolor()"),
				new AssistInfo(FUNCTION, "fontsizr(size) - String", "fontsize()"),
				new AssistInfo(FUNCTION, "anchor(size) - String", "anchor()"),
				new AssistInfo(FUNCTION, "link(size) - String", "link()"),
		};
		STATIC_MEMBERS.put("String", string);
	}
	
	private List<IAdditionalJavaScriptCompleter> completers = new ArrayList<IAdditionalJavaScriptCompleter>();
	private List<AssistInfo> functions = new ArrayList<AssistInfo>();
	
	/**
	 * Returns source code to parse from <code>ITextViewer</code>.
	 * <p>
	 * If you want to use this class with the your own editor 
	 * which supports the document contains JavaScript such as HTML/JSP,
	 * override this method as returns only JavaScript code.
	 * 
	 * @param viewer <code>ITextViewer</code>
	 * @return JavaScript source code
	 */
	protected String getSource(ITextViewer viewer){
		return viewer.getDocument().get();
	}
	
	/**
	 * Returns <code>AssistInfo[]</code> from <code>STATIC_MEMBERS</code> 
	 * and <code>IAdditionalJavaScriptCompleter</code>s.
	 */
	private AssistInfo[] getStatisMembers(String object){
		List<AssistInfo> result = new ArrayList<AssistInfo>();
		AssistInfo[] info = (AssistInfo[])STATIC_MEMBERS.get(object);
		if(info!=null){
			for(int i=0;i<info.length;i++){
				result.add(info[i]);
			}
		}
		if(object.equals("Object")){
			AssistInfo[] elementInfo = (AssistInfo[])STATIC_MEMBERS.get("Element");
			for(int i=0;i<elementInfo.length;i++){
				result.add(elementInfo[i]);
			}
		}
		for(int i=0;i<completers.size();i++){
			IAdditionalJavaScriptCompleter completer = completers.get(i);
			AssistInfo[] additionalInfo = (AssistInfo[]) completer.getClassAssistInformations().get(object);
			if(additionalInfo != null){
				for(int j=0;j<additionalInfo.length;j++){
					result.add(additionalInfo[j]);
				}
			}
		}
		if(result.size()==0){
			return null;
		}
		return result.toArray(new AssistInfo[result.size()]);
	}
	
	public ICompletionProposal[] computeCompletionProposals(ITextViewer viewer, int offset) {
		List<ICompletionProposal> proposal = new ArrayList<ICompletionProposal>();
		String source = getSource(viewer);
		
		String[] words = getLastWord(viewer, offset);
		String last = words[0];
		String word = words[1];
		
		List<String> addedStrings = new ArrayList<String>();
		
		if(last.endsWith(".")){
			String objName = last.substring(0, last.length()-1);
			
			JavaScriptModel model = new JavaScriptModel(source);
			JavaScriptContext context = model.getContextFromOffset(offset);
			if(context!=null){
				for(JavaScriptElement element: context.getVisibleElements()){
					if(element instanceof JavaScriptVariable && 
							((JavaScriptVariable) element).getName().equals(objName)){
						String type = ((JavaScriptVariable) element).getType();
						if(!type.equals("Object")){
							objName = type;
						}
						break;
					}
				}
			}
			
			AssistInfo[] info = getStatisMembers(objName);
			if(info==null && !isNumeric(objName)){
				info = getStatisMembers("Object");
			}
			if(info!=null){
				for(int i=0;i<info.length;i++){
					if(info[i].replaceString.startsWith(word)){
						proposal.add(info[i].createCompletionProposal(offset, word));
					}
				}
			}
		} else {
			for(int i=0;i<staticAssistInfo.size();i++){
				AssistInfo info = staticAssistInfo.get(i);
				if(info.replaceString.startsWith(word)){
					proposal.add(info.createCompletionProposal(offset, word));
				}
			}
			for(int i=0;i<completers.size();i++){
				IAdditionalJavaScriptCompleter completer = (IAdditionalJavaScriptCompleter) completers.get(i);
				AssistInfo[] prototypeAssistInfo = completer.getGlobalAssistInformations();
				for(int j=0;j<prototypeAssistInfo.length;j++){
					if(prototypeAssistInfo[j].replaceString.startsWith(word)){
						proposal.add(prototypeAssistInfo[j].createCompletionProposal(offset, word));
					}
				}
			}
			
			JavaScriptModel model = new JavaScriptModel(source);
			JavaScriptContext context = model.getContextFromOffset(offset);
			if(context!=null){
				JavaScriptElement[] children = context.getVisibleElements();
				for(int i=0;i<children.length;i++){
					if(children[i].getName().startsWith(word)){
						int type = UNDEF;
						int position = 0;
						String replace = null;
						
						if(children[i] instanceof JavaScriptFunction){
							type = FUNCTION;
							replace = children[i].getName() + "()";
							if(((JavaScriptFunction)children[i]).getArguments().length()==0){
								position = replace.length();
							} else{
								position = replace.length() - 1;
							}
						} else if(children[i] instanceof JavaScriptVariable){
							type = VARIABLE;
							replace = children[i].getName();
							position = replace.length();
						}
						if(!addedStrings.contains(replace)){
							proposal.add(new CompletionProposal(
									replace, offset - word.length(), word.length(),
									position, getImageFromType(type),
									children[i].toString(), null, null));
							addedStrings.add(replace);
						}
					}
				}
			}
		}
		
		for(AssistInfo info: functions){
			if(info.replaceString.startsWith(word) && !addedStrings.contains(info.replaceString)){
				proposal.add(info.createCompletionProposal(offset, word));
				addedStrings.add(info.replaceString);
			}
		}
		
		for(ICompletionProposal template: super.computeCompletionProposals(viewer, offset)){
			proposal.add(template);
		}
		
		// sort
		HTMLUtil.sortCompilationProposal(proposal);
		ICompletionProposal[] prop = proposal.toArray(new ICompletionProposal[proposal.size()]);
		
		return prop;
	}
	
	/**
	 * Returns <code>Image</code> from the assist proposal type.
	 * 
	 * @param type KEYWORD, VARIABLE or FUNCTION
	 * @return Image from the type
	 */
	private static Image getImageFromType(int type){
		switch(type){
			case KEYWORD:
				return HTMLPlugin.getDefault().getImageRegistry().get(HTMLPlugin.ICON_VALUE);
			case VARIABLE:
				return HTMLPlugin.getDefault().getImageRegistry().get(HTMLPlugin.ICON_VARIABLE);
			case FUNCTION:
				return HTMLPlugin.getDefault().getImageRegistry().get(HTMLPlugin.ICON_FUNCTION);
			case OBJECT:
				return HTMLPlugin.getDefault().getImageRegistry().get(HTMLPlugin.ICON_CLASS);
			default:
				return null;
		}
	}
	
	private boolean isNumeric(String value){
		for(int i=0;i<value.length();i++){
			char c = value.charAt(i);
			if(c < '0' || c > '9'){
				return false;
			}
		}
		return true;
	}
	
	/**
	 * Cuts out the last word of caret position.
	 * 
	 * @param viewer ITextViewer
	 * @param offset the caret offset
	 * @return the last word of caret position
	 */
	protected String[] getLastWord(ITextViewer viewer, int offset){
		String source = viewer.getDocument().get();
		source = source.substring(0, offset);
		String last = "";
		
		int start = source.lastIndexOf('\n');
		if(start==-1){
			start = source.lastIndexOf('\r');
		}
		if(start==-1){
			start = 0;
		}
		
		StringBuffer sb = new StringBuffer();
		for(int i=start;i < offset;i++){
			char c = source.charAt(i);
			if(Character.isWhitespace(c) || !Character.isJavaIdentifierPart(c)){
				if(sb.length()!=0){
					if(c=='.'){
						sb.append(c);
					}
					last = sb.toString();
					sb.setLength(0);
				}
			} else {
				sb.append(c);
			}
		}
		return new String[]{last, sb.toString()};
	}
	
	/**
	 * The structure for assist informations.
	 */
	public static class AssistInfo {
		
		private int type;
		private int position;
		private String displayString;
		private String replaceString;
		
		public AssistInfo(int type, String string){
			this(type, string, string);
		}
		
		public AssistInfo(int type, String string, int position){
			this(type, string, string, position);
		}
		
		public AssistInfo(int type, String displayString, String replaceString){
			this(type, displayString, replaceString, 
					replaceString.endsWith("()") ? replaceString.length()-1 : replaceString.length());
		}
		
		public AssistInfo(int type, String displayString, String replaceString, int position){
			this.type = type;
			this.displayString = displayString;
			this.replaceString = replaceString;
			this.position = position;
		}
		
		public CompletionProposal createCompletionProposal(int offset, String word){
			return new CompletionProposal(
					replaceString, offset - word.length(), word.length(),
					position, getImageFromType(type),
					displayString, null, null);
		}
	}
	
	public IContextInformation[] computeContextInformation(ITextViewer viewer, int offset) {
		return new ContextInformation[0];
	}

	public char[] getCompletionProposalAutoActivationCharacters() {
		return new char[0];
	}

	public char[] getContextInformationAutoActivationCharacters() {
		return new char[0];
	}

	public String getErrorMessage() {
		return "error";
	}

	public IContextInformationValidator getContextInformationValidator() {
		return new ContextInformationValidator(this);
	}
	
	protected TemplateContextType getContextType(ITextViewer viewer, IRegion region) {
		HTMLTemplateManager manager = HTMLTemplateManager.getInstance();
		return manager.getContextTypeRegistry().getContextType(JavaScriptContextType.CONTEXT_TYPE);
	}
	
	/**
	 * Updates internal informations.
	 * 
	 * @param file the editing file
	 */
	public void update(IFile file){
		try {
			HTMLProjectParams params = new HTMLProjectParams(file.getProject());
			
			String[] names = params.getJavaScriptCompleters();
			completers.clear();
			for(int i=0;i<names.length;i++){
				IAdditionalJavaScriptCompleter completer =
					AdditionalJavaScriptCompleterManager.getAdditionalJavaSCriptCompleter(names[i]);
				completers.add(completer);
			}
			
			String[] javaScripts = params.getJavaScripts();
			IWorkspaceRoot wsroot = ResourcesPlugin.getWorkspace().getRoot();
			functions.clear();
			
			LOOP: for(int i=0;i<javaScripts.length;i++){
				// Skip if completion proposal are provided by the additional completer
				for(int j=0;j<completers.size();j++){
					IAdditionalJavaScriptCompleter completer = (IAdditionalJavaScriptCompleter) completers.get(j);
					if(completer.isLibraryScriptFile(javaScripts[i])){
						continue LOOP;
					}
				}
				
				InputStream in = null;
				if(javaScripts[i].startsWith(JavaScriptLibraryTable.PREFIX)){
					IResource resource = wsroot.findMember(javaScripts[i].substring(JavaScriptLibraryTable.PREFIX.length()));
					if(resource!=null && resource instanceof IFile && resource.exists()){
						in = ((IFile)resource).getContents();
					}
				} else {
					in = new FileInputStream(javaScripts[i]);
				}
				String source = new String(IOUtil.readStream(in));
				JavaScriptModel model = new JavaScriptModel(source);
				JavaScriptElement[] elements = model.getChildren();
				
				for(int j=0;j<elements.length;j++){
					if(elements[j] instanceof JavaScriptFunction){
						String replace = elements[j].getName() + "()";
						int position = 0;
						if(((JavaScriptFunction)elements[j]).getArguments().length()==0){
							position = replace.length();
						} else{
							position = replace.length() - 1;
						}
						functions.add(new AssistInfo(
								FUNCTION, elements[j].toString(), replace, position));
					}
				}
			}
		} catch(Exception ex){
			HTMLPlugin.logException(ex);
		}
	}
	
}
